__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "arenas.guerrero.julian@outlook.com"


import errno
import os
import logging

from configparser import ConfigParser

from .constants import *
from .utils import get_valid_dir_path, get_valid_file_name, get_valid_file_path, remove_file_extension


CONFIGURATION_SECTION = 'CONFIGURATION'


##############################################################################
########################   CONFIGURATION PARAMETERS   ########################
##############################################################################

NA_FILTER = 'na_filter'
NA_VALUES = 'na_values'

OUTPUT_DIR = 'output_dir'
OUTPUT_FILE = 'output_file'
OUTPUT_FORMAT = 'output_format'
CLEAN_OUTPUT_DIR = 'clean_output_dir'
ONLY_PRINTABLE_CHARACTERS = 'only_printable_characters'

FILE_PATH = 'file_path'

MAPPING_PARTITION = 'mapping_partition'
INFER_SQL_DATATYPES = 'infer_sql_datatypes'
ENFORCE_SQL_QUERY_FILTER_NULL = 'enforce_sql_filter_null'

CHUNKSIZE = 'chunksize'

NUMBER_OF_PROCESSES = 'number_of_processes'

LOGGING_LEVEL = 'logging_level'
LOGGING_FILE = 'logging_file'

ORACLE_CLIENT_LIB_DIR = 'oracle_client_lib_dir'
ORACLE_CLIENT_CONFIG_DIR = 'oracle_client_config_dir'

READ_PARSED_MAPPINGS_PATH = 'read_parsed_mappings_path'
WRITE_PARSED_MAPPINGS_PATH = 'write_parsed_mappings_path'


##############################################################################
#########################   DATA SOURCE PARAMETERS   #########################
##############################################################################

MAPPINGS = 'mappings'
DATABASE_URL = 'db_url'


##############################################################################
########################   PARAMETERS DEFAULT VALUES   #######################
##############################################################################

# input parameters that are not to be completed with default value if they are empty
CONFIGURATION_OPTIONS_EMPTY_VALID = {
            OUTPUT_FILE: DEFAULT_OUTPUT_FILE,
            NA_VALUES: DEFAULT_NA_VALUES,
            READ_PARSED_MAPPINGS_PATH: DEFAULT_READ_PARSED_MAPPINGS_PATH,
            WRITE_PARSED_MAPPINGS_PATH: DEFAULT_WRITE_PARSED_MAPPINGS_PATH,
            MAPPING_PARTITION: PARTIAL_AGGREGATIONS_PARTITIONING,
            LOGGING_FILE: DEFAULT_LOGGING_FILE,
            ORACLE_CLIENT_LIB_DIR: DEFAULT_ORACLE_CLIENT_LIB_DIR,
            ORACLE_CLIENT_CONFIG_DIR: DEFAULT_ORACLE_CLIENT_LIB_DIR,
        }


# input parameters that are to be replaced with the default vale if they are empty
CONFIGURATION_OPTIONS_EMPTY_NON_VALID = {
            OUTPUT_DIR: DEFAULT_OUTPUT_DIR,
            OUTPUT_FORMAT: DEFAULT_OUTPUT_FORMAT,
            CLEAN_OUTPUT_DIR: DEFAULT_CLEAN_OUTPUT_DIR,
            ONLY_PRINTABLE_CHARACTERS: DEFAULT_ONLY_PRINTABLE_CHARACTERS,
            INFER_SQL_DATATYPES: DEFAULT_INFER_SQL_DATATYPES,
            CHUNKSIZE: DEFAULT_CHUNKSIZE,
            LOGGING_LEVEL: DEFAULT_LOGGING_LEVEL,
            NA_FILTER: DEFAULT_NA_FILTER,
            NUMBER_OF_PROCESSES: DEFAULT_NUMBER_OF_PROCESSES
        }


def _is_option_provided(config, option, empty_value_is_valid=False):
    """
    Checks whether a parameter is provided in the config. If empty value is not valid then the option will be considered
    as not provided if no value is provided for the option.
    """

    if not config.has_configuration_option(option):
        return False
    elif (config.get_configuration_option(option) == '') and (empty_value_is_valid is False):
        return False
    return True


class Config(ConfigParser):

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)

        self.configuration_section = CONFIGURATION_SECTION

    def complete_configuration_with_defaults(self):
        """
        Sets the parameters that are not specified in the config file to their corresponding default values.
        If there is no CONFIGURATION section in the config file it is also added.
        """

        # add configuration section if its does not exist
        if not self.has_section(self.configuration_section):
            self.add_section(self.configuration_section)

        # set default values for parameters that accept empty values, i.e. if the parameter is provided but no value
        # is provided the default value is not set, if the parameter itself is not provided the default parameter is set
        for configuration_option, configuration_option_default in CONFIGURATION_OPTIONS_EMPTY_VALID.items():
            if not _is_option_provided(self, configuration_option, empty_value_is_valid=True):
                self.set(self.configuration_section, configuration_option, str(configuration_option_default))
        # set default values for parameters that do not accept empty values, i.e. if the parameter is provided but no
        # value is provided the default value is set, if the parameter itself is not provided the default parameter is
        # also set
        for configuration_option, configuration_option_default in CONFIGURATION_OPTIONS_EMPTY_NON_VALID.items():
            if not _is_option_provided(self, configuration_option):
                self.set(self.configuration_section, configuration_option, str(configuration_option_default))

    def validate_configuration_section(self):
        # OUTPUT DIR & FILE, WRITE PARSED MAPPINGS FILE, LOGS FILE
        self.set_output_dir(get_valid_dir_path(self.get_output_dir()))
        self.set_output_file(get_valid_file_name(self.get_output_file()))
        self.set_parsed_mappings_write_path(get_valid_file_path(self.get_parsed_mappings_write_path()))
        self.set_logging_file(get_valid_file_path(self.get_logging_file()))

        # OUTPUT FORMAT
        output_format = str(self.get_output_format()).upper()
        self.set_output_format(output_format)
        if output_format not in VALID_OUTPUT_FORMATS:
            raise ValueError(OUTPUT_FORMAT + ' value `' + self.get_output_format() +
                             '` is not valid. It must be in: ' + str(VALID_OUTPUT_FORMATS) + '.')

        # LOGGING LEVEL
        logging_level = str(self.get_logging_level()).upper()
        self.set_logging_level(logging_level)
        if logging_level not in VALID_LOGGING_LEVEL:
            raise ValueError(LOGGING_LEVEL + ' value `' + self.get_logging_level() +
                             '` is not valid. It must be in: ' + str(VALID_LOGGING_LEVEL) + '.')

        # MAPPING PARTITIONING
        mapping_partitioning = str(self.get_mapping_partition()).upper()
        self.set_mapping_partition(mapping_partitioning)
        if mapping_partitioning not in NO_PARTITIONING + [PARTIAL_AGGREGATIONS_PARTITIONING] + [
                MAXIMAL_PARTITIONING]:
            raise ValueError(
                MAPPING_PARTITION + ' value `' + self.get_mapping_partition() + '` is not valid. It must be in: ' + str(
                    [MAXIMAL_PARTITIONING] + [
                        PARTIAL_AGGREGATIONS_PARTITIONING] + NO_PARTITIONING) + '.')

    def log_config_info(self):
        logging.debug('CONFIGURATION: ' + str(dict(self.items(self.configuration_section))))

        for data_source_section in self.get_data_sources_sections():
            logging.debug('DATA SOURCE `' + data_source_section + '`: ' + str(dict(self.items(data_source_section))))

    #################################################################################
    #######################   CONFIGURATION SECTION METHODS   #######################
    #################################################################################

    def has_configuration_option(self, option):
        return self.has_option(self.configuration_section, option)

    def has_multiple_data_sources(self):
        return len(self.get_data_sources_sections()) > 1

    def is_multiprocessing_enabled(self):
        return self.getint(self.configuration_section, NUMBER_OF_PROCESSES) > 1

    def is_read_parsed_mappings_file_provided(self):
        if self.get(self.configuration_section, READ_PARSED_MAPPINGS_PATH):
            return True
        else:
            return False

    def is_write_parsed_mappings_file_provided(self):
        if self.get(self.configuration_section, WRITE_PARSED_MAPPINGS_PATH):
            return True
        else:
            return False

    def is_oracle_client_lib_dir_provided(self):
        if self.get(self.configuration_section, ORACLE_CLIENT_LIB_DIR):
            return True
        else:
            return False

    def is_oracle_client_config_dir_provided(self):
        if self.get(self.configuration_section, ORACLE_CLIENT_CONFIG_DIR):
            return True
        else:
            return False

    def infer_sql_datatypes(self):
        return self.getboolean(self.configuration_section, INFER_SQL_DATATYPES)

    def enforce_sql_filter_null(self):
        return self.getboolean(self.configuration_section, ENFORCE_SQL_QUERY_FILTER_NULL)

    def clean_output_dir(self):
        return self.getboolean(self.configuration_section, CLEAN_OUTPUT_DIR)

    def only_write_printable_characters(self):
        return self.getboolean(self.configuration_section, ONLY_PRINTABLE_CHARACTERS)

    def get_configuration_option(self, option):
        return self.get(self.configuration_section, option)

    def get_number_of_processes(self):
        return self.getint(self.configuration_section, NUMBER_OF_PROCESSES)

    def get_logging_level(self):
        return self.get(self.configuration_section, LOGGING_LEVEL)

    def get_logging_file(self):
        return self.get(self.configuration_section, LOGGING_FILE)

    def get_parsed_mappings_read_path(self):
        return self.get(self.configuration_section, READ_PARSED_MAPPINGS_PATH)

    def get_parsed_mappings_write_path(self):
        return self.get(self.configuration_section, WRITE_PARSED_MAPPINGS_PATH)

    def get_oracle_client_lib_dir(self):
        return self.get(self.configuration_section, ORACLE_CLIENT_LIB_DIR)

    def get_oracle_client_config_dir(self):
        return self.get(self.configuration_section, ORACLE_CLIENT_CONFIG_DIR)

    def get_mapping_partition(self):
        return self.get(self.configuration_section, MAPPING_PARTITION)

    def get_output_dir(self):
        return self.get(self.configuration_section, OUTPUT_DIR)

    def get_output_file(self):
        return self.get(self.configuration_section, OUTPUT_FILE)

    def get_output_format(self):
        return self.get(self.configuration_section, OUTPUT_FORMAT)

    def get_chunksize(self):
        return self.getint(self.configuration_section, CHUNKSIZE)

    def apply_na_filter(self):
        return self.getboolean(self.configuration_section, NA_FILTER)

    def get_na_values(self):
        if self.apply_na_filter():
            return list(set(self.get(self.configuration_section, NA_VALUES).split(',')))
        else:
            return []

    def get_output_file_path(self, mapping_partition=None):
        if self.get_output_file():
            file_path = os.path.join(self.get_output_dir(), self.get_output_file())
            # remove file extension, we will set it based on the output format
            file_path = remove_file_extension(file_path)
        elif mapping_partition:
            file_path = os.path.join(self.get_output_dir(), mapping_partition)
        else:
            # neither output_file was specified nor mapping partition are used. Use default output_file.
            file_path = os.path.join(self.get_output_dir(), OUTPUT_FILE)

        # add file extension
        file_path += OUTPUT_FORMAT_FILE_EXTENSION[self.get_output_format()]

        return file_path

    def set_mapping_partition(self, mapping_partitioning):
        self.set(self.configuration_section, MAPPING_PARTITION, mapping_partitioning.upper())

    def set_output_dir(self, output_dir):
        self.set(self.configuration_section, OUTPUT_DIR, output_dir)

    def set_output_file(self, output_file):
        self.set(self.configuration_section, OUTPUT_FILE, output_file)

    def set_parsed_mappings_write_path(self, parsed_mappings_write_path):
        self.set(self.configuration_section, WRITE_PARSED_MAPPINGS_PATH, parsed_mappings_write_path)

    def set_logging_file(self, logging_file):
        self.set(self.configuration_section, LOGGING_FILE, logging_file)

    def set_logging_level(self, logging_level):
        self.set(self.configuration_section, LOGGING_LEVEL, logging_level)

    def set_output_format(self, output_format):
        self.set(self.configuration_section, OUTPUT_FORMAT, output_format)

    ################################################################################
    #######################   DATA SOURCE SECTIONS METHODS   #######################
    ################################################################################

    def get_data_sources_sections(self):
        return list(set(self.sections()) - {self.configuration_section})

    def has_file_path(self, source_section):
        return self.has_option(source_section, FILE_PATH)

    def get_file_path(self, source_section):
        return self.get(source_section, FILE_PATH)

    def get_mappings_files(self, source_section):
        mapping_file_paths = []

        for mapping_path in self.get(source_section, MAPPINGS).split(','):
            # if it is a file load the mapping triples to the graph
            if os.path.isfile(mapping_path):
                mapping_file_paths.append(mapping_path)
            # if it is a directory process all the mapping files within the root of the directory
            elif os.path.isdir(mapping_path):
                for mapping_file_name in os.listdir(mapping_path):
                    mapping_file = os.path.join(mapping_path, mapping_file_name)
                    if os.path.isfile(mapping_file):
                        mapping_file_paths.append(mapping_file)
            else:
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), mapping_path)

        return mapping_file_paths

    def get_database_url(self, source_section):
        return self.get(source_section, DATABASE_URL)

    def has_database_url(self, source_section):
        return self.has_option(source_section, DATABASE_URL)



__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "arenas.guerrero.julian@outlook.com"


import json
import pandas as pd

from lxml import etree
from jsonpath import JSONPath

from ..constants import *


def get_file_data(config, mapping_rule, references):
    file_source_type = mapping_rule['source_type']

    if file_source_type in [CSV, TSV]:
        return _read_csv(config, mapping_rule, references, file_source_type)
    elif file_source_type in EXCEL:
        return _read_excel(config, mapping_rule, references)
    elif file_source_type == PARQUET:
        return _read_parquet(mapping_rule, references)
    elif file_source_type in FEATHER:
        return _read_feather(mapping_rule, references)
    elif file_source_type == ORC:
        return _read_orc(mapping_rule, references)
    elif file_source_type == STATA:
        return _read_stata(config, mapping_rule, references)
    elif file_source_type in SAS:
        return _read_sas(config, mapping_rule, references)
    elif file_source_type == SPSS:
        return _read_spss(mapping_rule, references)
    elif file_source_type == JSON:
        return _read_json(mapping_rule, references)
    elif file_source_type == XML:
        return _read_xml(mapping_rule, references)
    else:
        raise ValueError('Found an invalid source type. Found value `' + file_source_type + '`.')


def _read_csv(config, mapping_rule, references, file_source_type):
    delimiter = ',' if file_source_type == 'CSV' else '\t'

    return pd.read_table(str(mapping_rule['data_source']),
                         delimiter=delimiter,
                         index_col=False,
                         encoding='utf-8',
                         encoding_errors='strict',
                         usecols=references,
                         chunksize=config.get_chunksize(),
                         engine='c',
                         dtype=str,
                         keep_default_na=False,
                         na_values=config.get_na_values(),
                         na_filter=False)


def _read_parquet(mapping_rule, references):
    parquet_df = pd.read_parquet(str(mapping_rule['data_source']),
                                 engine='pyarrow',
                                 columns=references)

    return [parquet_df]     # return as list because it does not support chunksize


def _read_feather(mapping_rule, references):
    feather_df = pd.read_feather(str(mapping_rule['data_source']),
                                 use_threads=True,
                                 columns=references)

    return [feather_df]


def _read_orc(mapping_rule, references):
    orc_df = pd.read_orc(str(mapping_rule['data_source']),
                         encoding='utf-8',
                         columns=references)

    return [orc_df]


def _read_stata(config, mapping_rule, references):
    return pd.read_stata(str(mapping_rule['data_source']),
                         columns=references,
                         chunksize=config.get_chunksize(),
                         convert_dates=False,
                         convert_categoricals=False,
                         convert_missing=False,
                         preserve_dtypes=False,
                         order_categoricals=False)


def _read_sas(config, mapping_rule, references):
    sas_df = pd.read_sas(str(mapping_rule['data_source']),
                         encoding='utf-8',
                         chunksize=config.get_chunksize())
    sas_df = sas_df[references]

    return sas_df


def _read_spss(mapping_rule, references):
    spss_df = pd.read_spss(str(mapping_rule['data_source']),
                           usecols=references,
                           convert_categoricals=False)

    return [spss_df]


def _read_excel(config, mapping_rule, references):
    excel_df = pd.read_excel(str(mapping_rule['data_source']),
                             sheet_name=0,
                             engine='openpyxl',
                             usecols=references,
                             dtype=str,
                             keep_default_na=False,
                             na_values=config.get_na_values(),
                             na_filter=False)

    return [excel_df]


def _read_json(mapping_rule, references):
    with open(str(mapping_rule['data_source']), encoding='utf-8') as json_file:
        json_data = json.load(json_file)

    jsonpath_expression = mapping_rule['iterator'] + '.('
    for reference in references:
        jsonpath_expression += reference + ','
    jsonpath_expression = jsonpath_expression[:-1] + ')'

    jsonpath_result = JSONPath(jsonpath_expression).parse(json_data)
    json_df = pd.DataFrame.from_records(jsonpath_result)
    json_df.dropna(axis=0, how='any', inplace=True)

    # in case lists where retrieved, explode them e.g.: ['a', 'b'] -> 'a', 'b'
    for reference in references:
        json_df = json_df.explode(reference)

    return [json_df]


def _read_xml(mapping_rule, references):
    with open(str(mapping_rule['data_source']), encoding='utf-8') as xml_file:
        xml_root = etree.parse(xml_file).getroot()

    xpath_expression = mapping_rule['iterator']
    for reference in references:
        xpath_expression += '[' + reference + ']'

    xpath_result = xml_root.xpath(xpath_expression)
    xpath_result = [[e.find(reference).text for reference in references] for e in xpath_result]

    xml_df = pd.DataFrame.from_records(xpath_result, columns=references)

    return [xml_df]



__author__ = "Julián Arenas-Guerrero"
__copyright__ = "Copyright (C) 2020-2021 Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "arenas.guerrero.julian@outlook.com"


import multiprocessing as mp


##############################################################################
########################   MAPPING PARTITION OPTIONS   #######################
##############################################################################

PARTIAL_AGGREGATIONS_PARTITIONING = 'PARTIAL-AGGREGATIONS'
MAXIMAL_PARTITIONING = 'MAXIMAL'
NO_PARTITIONING = ['NO', 'FALSE', 'OFF', '0']


##############################################################################
#########################   DATA SOURCE TYPES   ##############################
##############################################################################

RDB = 'RDB'
CSV = 'CSV'
TSV = 'TSV'
EXCEL = ['XLS', 'XLSX', 'XLSM', 'XLSB', 'ODF', 'ODS', 'ODT']
PARQUET = 'PARQUET'
FEATHER = ['FEATHER', 'FEA']
ORC = 'ORC'
STATA = 'DTA'
SAS = ['XPT', 'SAS7BDAT']
SPSS = 'SAV'
JSON = 'JSON'
XML = 'XML'

# DBMSs
MYSQL = 'MYSQL'
MARIADB = 'MARIADB'
MSSQL = 'MSSQL'
ORACLE = 'ORACLE'
POSTGRESQL = 'POSTGRESQL'
SQLITE = 'SQLITE'

FILE_SOURCE_TYPES = [CSV, TSV, PARQUET, ORC, STATA, SPSS, JSON, XML] + EXCEL + FEATHER + SAS
DATA_SOURCE_TYPES = [RDB] + FILE_SOURCE_TYPES

# RDF serializations
NTRIPLES = 'N-TRIPLES'
NQUADS = 'N-QUADS'


##############################################################################
#########################   VALID ARGUMENTS VALUES   #########################
##############################################################################

VALID_OUTPUT_FORMATS = ['N-TRIPLES', NQUADS]
VALID_PROCESS_START_METHOD = ['DEFAULT', 'SPAWN', 'FORK', 'FORKSERVER']
VALID_LOGGING_LEVEL = ['NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']


##############################################################################
###################   FILE EXTENSIONS FOR OUTPUT FORMATS   ###################
##############################################################################

OUTPUT_FORMAT_FILE_EXTENSION = {
    NTRIPLES: '.nt',
    NQUADS: '.nq'
}


##############################################################################
###########################   R2RML SPECIFICATION   ##########################
##############################################################################

R2RML_NAMESPACE = 'http://www.w3.org/ns/r2rml#'

# classes
R2RML_BASE_TABLE_OR_VIEW_CLASS = 'http://www.w3.org/ns/r2rml#BaseTableOrView'
R2RML_GRAPH_MAP_CLASS = 'http://www.w3.org/ns/r2rml#GraphMap'
R2RML_JOIN_CLASS = 'http://www.w3.org/ns/r2rml#Join'
R2RML_LOGICAL_TABLE_CLASS = 'http://www.w3.org/ns/r2rml#LogicalTable'
R2RML_OBJECT_MAP_CLASS = 'http://www.w3.org/ns/r2rml#ObjectMap'
R2RML_PREDICATE_MAP_CLASS = 'http://www.w3.org/ns/r2rml#PredicateMap'
R2RML_PREDICATE_OBJECT_MAP_CLASS = 'http://www.w3.org/ns/r2rml#PredicateObjectMap'
R2RML_R2RML_VIEW_CLASS = 'http://www.w3.org/ns/r2rml#R2RMLView'
R2RML_REF_OBJECT_MAP_CLASS = 'http://www.w3.org/ns/r2rml#RefObjectMap'
R2RML_SUBJECT_MAP_CLASS = 'http://www.w3.org/ns/r2rml#SubjectMap'
R2RML_TERM_MAP_CLASS = 'http://www.w3.org/ns/r2rml#TermMap'
R2RML_TRIPLES_MAP_CLASS = 'http://www.w3.org/ns/r2rml#TriplesMap'

# properties
R2RML_LOGICAL_TABLE = 'http://www.w3.org/ns/r2rml#logicalTable'
R2RML_TABLE_NAME = 'http://www.w3.org/ns/r2rml#tableName'
R2RML_PARENT_TRIPLES_MAP = 'http://www.w3.org/ns/r2rml#parentTriplesMap'
R2RML_SUBJECT_MAP = 'http://www.w3.org/ns/r2rml#subjectMap'
R2RML_PREDICATE_MAP = 'http://www.w3.org/ns/r2rml#predicateMap'
R2RML_OBJECT_MAP = 'http://www.w3.org/ns/r2rml#objectMap'
R2RML_GRAPH_MAP = 'http://www.w3.org/ns/r2rml#graphMap'
R2RML_SUBJECT_CONSTANT_SHORTCUT = 'http://www.w3.org/ns/r2rml#subject'
R2RML_PREDICATE_CONSTANT_SHORTCUT = 'http://www.w3.org/ns/r2rml#predicate'
R2RML_OBJECT_CONSTANT_SHORTCUT = 'http://www.w3.org/ns/r2rml#object'
R2RML_GRAPH_CONSTANT_SHORTCUT = 'http://www.w3.org/ns/r2rml#graph'
R2RML_PREDICATE_OBJECT_MAP = 'http://www.w3.org/ns/r2rml#predicateObjectMap'
R2RML_CONSTANT = 'http://www.w3.org/ns/r2rml#constant'
R2RML_TEMPLATE = 'http://www.w3.org/ns/r2rml#template'
R2RML_COLUMN = 'http://www.w3.org/ns/r2rml#column'
R2RML_CLASS = 'http://www.w3.org/ns/r2rml#class'
R2RML_CHILD = 'http://www.w3.org/ns/r2rml#child'
R2RML_PARENT = 'http://www.w3.org/ns/r2rml#parent'
R2RML_JOIN_CONDITION = 'http://www.w3.org/ns/r2rml#joinCondition'
R2RML_DATATYPE = 'http://www.w3.org/ns/r2rml#datatype'
R2RML_LANGUAGE = 'http://www.w3.org/ns/r2rml#language'
R2RML_SQL_QUERY = 'http://www.w3.org/ns/r2rml#sqlQuery'
R2RML_SQL_VERSION = 'http://www.w3.org/ns/r2rml#sqlVersion'
R2RML_TERM_TYPE = 'http://www.w3.org/ns/r2rml#termType'

# other
R2RML_DEFAULT_GRAPH = 'http://www.w3.org/ns/r2rml#defaultGraph'
R2RML_IRI = 'http://www.w3.org/ns/r2rml#IRI'
R2RML_LITERAL = 'http://www.w3.org/ns/r2rml#Literal'
R2RML_BLANK_NODE = 'http://www.w3.org/ns/r2rml#BlankNode'
R2RML_SQL2008 = 'http://www.w3.org/ns/r2rml#SQL2008'


##############################################################################
############################   RML SPECIFICATION   ###########################
##############################################################################

QL_NAMESPACE = 'http://semweb.mmlab.be/ns/ql#'
QL_CSV = 'http://semweb.mmlab.be/ns/ql#CSV'
QL_JSON = 'http://semweb.mmlab.be/ns/ql#JSONPath'
QL_XML = 'http://semweb.mmlab.be/ns/ql#XPath'

RML_NAMESPACE = 'http://semweb.mmlab.be/ns/rml#'
RML_LOGICAL_SOURCE = 'http://semweb.mmlab.be/ns/rml#logicalSource'
RML_QUERY = 'http://semweb.mmlab.be/ns/rml#query'
RML_ITERATOR = 'http://semweb.mmlab.be/ns/rml#iterator'
RML_REFERENCE = 'http://semweb.mmlab.be/ns/rml#reference'
RML_REFERENCE_FORMULATION = 'http://semweb.mmlab.be/ns/rml#referenceFormulation'


##############################################################################
#############################   XSD DATA TYPES   #############################
##############################################################################
XSD_HEX_BINARY = 'http://www.w3.org/2001/XMLSchema#hexBinary'
XSD_INTEGER = 'http://www.w3.org/2001/XMLSchema#integer'
XSD_DECIMAL = 'http://www.w3.org/2001/XMLSchema#decimal'
XSD_DOUBLE = 'http://www.w3.org/2001/XMLSchema#double'
XSD_BOOLEAN = 'http://www.w3.org/2001/XMLSchema#boolean'
XSD_DATE = 'http://www.w3.org/2001/XMLSchema#date'
XSD_TIME = 'http://www.w3.org/2001/XMLSchema#time'
XSD_DATETIME = 'http://www.w3.org/2001/XMLSchema#dateTime'
XSD_STRING = 'http://www.w3.org/2001/XMLSchema#string'

##############################################################################
##################################   OTHER   #################################
##############################################################################

AUXILIAR_UNIQUE_REPLACING_STRING = 'zzyy_xxww\u200B'


##############################################################################
########################   ARGUMENTS DEFAULT VALUES   ########################
##############################################################################

DEFAULT_OUTPUT_DIR = 'output'
DEFAULT_OUTPUT_FILE = 'result'
DEFAULT_OUTPUT_FORMAT = NQUADS
DEFAULT_CLEAN_OUTPUT_DIR = 'yes'
DEFAULT_LOGGING_FILE = ''
DEFAULT_LOGGING_LEVEL = 'INFO'
DEFAULT_INFER_SQL_DATATYPES = 'no'
DEFAULT_NUMBER_OF_PROCESSES = 2 * mp.cpu_count()
DEFAULT_CHUNKSIZE = 100000
DEFAULT_NA_FILTER = 'yes'
DEFAULT_NA_VALUES = ',#N/A,N/A,#N/A N/A,n/a,NA,<NA>,#NA,NULL,null,NaN,nan,None'
DEFAULT_ONLY_PRINTABLE_CHARACTERS = 'no'

# ORACLE
DEFAULT_ORACLE_CLIENT_LIB_DIR = ''
DEFAULT_ORACLE_CLIENT_CONFIG_DIR = ''

# DEVELOPMENT OPTIONS
DEFAULT_READ_PARSED_MAPPINGS_PATH = ''
DEFAULT_WRITE_PARSED_MAPPINGS_PATH = ''


__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "arenas.guerrero.julian@outlook.com"


import logging
import pandas as pd
import sql_metadata

from sqlalchemy import create_engine
from sqlalchemy.pool import NullPool

from ..constants import *


# PostgresSQL data types: https://www.postgresql.org/docs/14/datatype.html
# Oracle data types: https://docs.oracle.com/cd/A58617_01/server.804/a58241/ch5.htm
# MySQL data types: https://dev.mysql.com/doc/refman/8.0/en/data-types.html
SQL_RDF_DATATYPE = {
    'BINARY': XSD_HEX_BINARY,
    'VARBINARY': XSD_HEX_BINARY,
    'BLOB': XSD_HEX_BINARY,
    'BFILE': XSD_HEX_BINARY,
    'RAW': XSD_HEX_BINARY,
    'LONG RAW': XSD_HEX_BINARY,

    'INTEGER': XSD_INTEGER,
    'INT': XSD_INTEGER,
    'SMALLINT': XSD_INTEGER,
    'INT8': XSD_INTEGER,
    'INT4': XSD_INTEGER,
    'BIGINT': XSD_INTEGER,
    'BIGSERIAL': XSD_INTEGER,
    'SMALLSERIAL': XSD_INTEGER,
    'INT2': XSD_INTEGER,
    'SERIAL2': XSD_INTEGER,
    'SERIAL4': XSD_INTEGER,
    'SERIAL8': XSD_INTEGER,

    'DECIMAL': XSD_DECIMAL,
    'NUMERIC': XSD_DECIMAL,

    'FLOAT': XSD_DOUBLE,
    'FLOAT8': XSD_DOUBLE,
    'REAL': XSD_DOUBLE,
    'DOUBLE': XSD_DOUBLE,
    'DOUBLE PRECISION': XSD_DOUBLE,
    'NUMBER': XSD_DOUBLE,

    'BOOL': XSD_BOOLEAN,
    'TINYINT': XSD_BOOLEAN,
    'BOOLEAN': XSD_BOOLEAN,

    'DATE': XSD_DATE,
    'TIME': XSD_TIME,
    'DATETIME': XSD_DATETIME,
    'TIMESTAMP': XSD_DATETIME
}


def _replace_query_enclosing_characters(sql_query, db_dialect):
    dialect_sql_query = ''

    if db_dialect in [MYSQL, MARIADB]:
        dialect_sql_query = sql_query   # the query already uses backticks as enclosed characters
    elif db_dialect == MSSQL:
        # replace backticks with square brackets
        square_brackets = ['[', ']']
        num_enclosing_char = 0
        for char in sql_query:
            if char == '`':
                dialect_sql_query = dialect_sql_query + square_brackets[num_enclosing_char % 2]
                num_enclosing_char += 1
            else:
                dialect_sql_query = dialect_sql_query + char
    else:
        # replace backticks with double quotes
        dialect_sql_query = sql_query.replace('`', '"')

    return dialect_sql_query


def _relational_db_connection(config, source_name):
    db_connection = create_engine(config.get_database_url(source_name), poolclass=NullPool)
    db_dialect = db_connection.dialect.name.upper()

    return db_connection, db_dialect


def _get_column_table_datatype(config, source_name, table_name, column_name):
    db_connection, db_dialect = _relational_db_connection(config, source_name)

    if db_dialect == ORACLE:
        # Oracle does not use the standard INFORMATION_SCHEMA
        sql_query = "SELECT t.data_type FROM all_tab_columns t WHERE t.TABLE_NAME = '" + table_name + \
                    "' AND t.COLUMN_NAME='" + column_name + "'"
    else:
        sql_query = "SELECT `data_type` FROM `information_schema`.`columns` WHERE `table_name`='" + table_name + \
                    "' AND `column_name`='" + column_name + "'"
        sql_query = _replace_query_enclosing_characters(sql_query, db_dialect)

    try:
        query_results_df = pd.read_sql(sql_query, con=db_connection)
    except:
        raise Exception('Query [' + sql_query + '] has failed to execute.')

    data_type = ''
    if 'data_type' in query_results_df.columns and len(query_results_df) == 1:
        data_type = query_results_df['data_type'][0]
    elif 'DATA_TYPE' in query_results_df.columns and len(query_results_df) == 1:
        data_type = query_results_df['DATA_TYPE'][0]

    if data_type.upper() in SQL_RDF_DATATYPE:
        return SQL_RDF_DATATYPE[data_type.upper()]
    else:
        return None


def get_rdb_reference_datatype(config, mapping_rule, reference):
    inferred_data_type = ''

    if pd.notna(mapping_rule['tablename']):
        inferred_data_type = _get_column_table_datatype(config, mapping_rule['source_name'],
                                                        mapping_rule['tablename'], reference)
    elif pd.notna(mapping_rule['query']):
        # if mapping rule has a query, get the table names in the query
        table_names = sql_metadata.Parser(mapping_rule['query']).tables
        for table_name in table_names:
            # for each table in the query get the datatype of the object reference in that table if an
            # exception is thrown, then the reference is not a column in that table, and nothing is done
            try:
                inferred_data_type = _get_column_table_datatype(config, mapping_rule['source_name'],
                                                                table_name, reference)
                if inferred_data_type:
                    # already found it, end looping
                    break
            except:
                pass

    return inferred_data_type


def _build_sql_query(mapping_rule, references):
    """
    Build a query for MYSQL using backticks '`' as enclosing character. This character will later be replaced with the
    one corresponding one to the dialect that applies.
    """

    if pd.notna(mapping_rule['query']):
        query = mapping_rule['query']
    elif len(references) > 0:
        query = 'SELECT '
        # query = query + 'DISTINCT ' # TODO: is this more efficient?
        for reference in references:
            query = query + '`' + reference + '`, '
        query = query[:-2] + ' FROM `' + mapping_rule['tablename'] + '` WHERE '
        for reference in references:
            query = query + '`' + reference + '` IS NOT NULL AND '
        query = query[:-5]
    else:
        query = None

    return query


def get_sql_data(config, mapping_rule, references):
    sql_query = _build_sql_query(mapping_rule, references)
    db_connection, db_dialect = _relational_db_connection(config, mapping_rule['source_name'])
    sql_query = _replace_query_enclosing_characters(sql_query, db_dialect)

    if sql_query is not None:
        logging.debug('SQL query for mapping rule `' + str(mapping_rule['id']) + '`: [' + sql_query + ']')

    result_chunks = pd.read_sql(sql_query,
                                con=db_connection,
                                chunksize=config.get_chunksize(),
                                coerce_float=False)

    return result_chunks


def setup_oracle(config):
    if config.is_oracle_client_config_dir_provided() or config.is_oracle_client_lib_dir_provided():
        import cx_Oracle

    if config.is_oracle_client_config_dir_provided() and config. is_oracle_client_lib_dir_provided():
        cx_Oracle.init_oracle_client(lib_dir=config.get_oracle_client_lib_dir(),
                                     config_dir=config.get_oracle_client_config_dir())
    elif config.is_oracle_client_config_dir_provided():
        cx_Oracle.init_oracle_client(config_dir=config.get_oracle_client_config_dir())
    elif config.is_oracle_client_lib_dir_provided():
        cx_Oracle.init_oracle_client(lib_dir=config.get_oracle_client_lib_dir())



__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "arenas.guerrero.julian@outlook.com"


import logging

from rdflib import Graph

from .engine import retrieve_mappings
from .args_parser import load_config_from_argument
from .materializer import _materialize_mapping_rule
from .utils import get_subject_maps
from .data_source.relational_database import setup_oracle


def materialize(config):

    config = load_config_from_argument(config)

    setup_oracle(config)

    mappings_df = retrieve_mappings(config)
    subject_maps_df = get_subject_maps(mappings_df)
    mapping_partitions = [group for _, group in mappings_df.groupby(by='mapping_partition')]

    num_triples = 0
    graph = Graph()
    for mapping_partition in mapping_partitions:
        triples = set()
        for i, mapping_rule in mapping_partition.iterrows():
            triples.update(set(_materialize_mapping_rule(mapping_rule, subject_maps_df, config)))

            logging.debug(str(len(triples)) + ' triples generated for mapping rule `' + str(mapping_rule['id']) + '`.')

        rdf_ntriples = '.\n'.join(triples)
        rdf_ntriples += '.'
        graph.parse(data=rdf_ntriples, format=config.get_output_format().replace('-', '').lower())

    logging.info('Number of triples generated in total: ' + str(num_triples) + '.')

    return graph



__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "arenas.guerrero.julian@outlook.com"

import pandas as pd
import numpy as np
import multiprocessing as mp
import rdflib
import logging

from ..constants import *
from ..utils import *
from ..mapping.mapping_constants import MAPPINGS_DATAFRAME_COLUMNS, MAPPING_PARSING_QUERY, JOIN_CONDITION_PARSING_QUERY
from ..mapping.mapping_partitioner import MappingPartitioner
from ..data_source.relational_database import get_rdb_reference_datatype


def _mapping_to_rml(mapping_graph):
    """
    Replaces R2RML rules in the graph with the corresponding RML rules.
    """

    # add namespaces
    mapping_graph.bind('rr', rdflib.term.URIRef(RML_NAMESPACE))
    mapping_graph.bind('rml', rdflib.term.URIRef(RML_NAMESPACE))
    mapping_graph.bind('ql', rdflib.term.URIRef(QL_NAMESPACE))

    # add reference formulation and sql version for RDB sources
    query = 'SELECT ?logical_source ?x WHERE { ?logical_source <' + R2RML_TABLE_NAME + '> ?x . } '
    for logical_source, _ in mapping_graph.query(query):
        mapping_graph.add((logical_source, rdflib.term.URIRef(R2RML_SQL_VERSION), rdflib.term.URIRef(R2RML_SQL2008)))
    query = 'SELECT ?logical_source ?x WHERE { ?logical_source <' + R2RML_SQL_QUERY + '> ?x . } '
    for logical_source, _ in mapping_graph.query(query):
        mapping_graph.add((logical_source, rdflib.term.URIRef(R2RML_SQL_VERSION), rdflib.term.URIRef(R2RML_SQL2008)))
        mapping_graph.add((logical_source, rdflib.term.URIRef(RML_REFERENCE_FORMULATION), rdflib.term.URIRef(QL_CSV)))

    # replace R2RML predicates with the equivalent RML predicates
    mapping_graph = replace_predicates_in_graph(mapping_graph, R2RML_LOGICAL_TABLE, RML_LOGICAL_SOURCE)
    mapping_graph = replace_predicates_in_graph(mapping_graph, R2RML_SQL_QUERY, RML_QUERY)
    mapping_graph = replace_predicates_in_graph(mapping_graph, R2RML_COLUMN, RML_REFERENCE)

    # remove R2RML classes
    mapping_graph.remove((None, rdflib.term.URIRef(R2RML_R2RML_VIEW_CLASS), None))
    mapping_graph.remove((None, rdflib.term.URIRef(R2RML_LOGICAL_TABLE_CLASS), None))

    return mapping_graph


def _expand_constant_shortcut_properties(mapping_graph):
    """
    Expand constant shortcut properties rr:subject, rr:predicate, rr:object and rr:graph.
    See R2RML specification (https://www.w3.org/2001/sw/rdb2rdf/r2rml/#constant).
    """

    constant_properties = [R2RML_SUBJECT_MAP, R2RML_PREDICATE_MAP, R2RML_OBJECT_MAP, R2RML_GRAPH_MAP]
    constant_shortcuts = [R2RML_SUBJECT_CONSTANT_SHORTCUT, R2RML_PREDICATE_CONSTANT_SHORTCUT,
                          R2RML_OBJECT_CONSTANT_SHORTCUT, R2RML_GRAPH_CONSTANT_SHORTCUT]

    for constant_property, constant_shortcut in zip(constant_properties, constant_shortcuts):
        for s, o in mapping_graph.query('SELECT ?s ?o WHERE {?s <' + constant_shortcut + '> ?o .}'):
            blanknode = rdflib.BNode()
            mapping_graph.add((s, rdflib.term.URIRef(constant_property), blanknode))
            mapping_graph.add((blanknode, rdflib.term.URIRef(R2RML_CONSTANT), o))

        mapping_graph.remove((None, rdflib.term.URIRef(constant_shortcut), None))

    return mapping_graph


def _rdf_class_to_pom(mapping_graph):
    """
    Replace rr:class definitions by predicate object maps.
    """

    query = 'SELECT ?tm ?c WHERE { ' \
            '?tm <' + R2RML_SUBJECT_MAP + '> ?sm . ' \
            '?sm <' + R2RML_CLASS + '> ?c . }'
    for tm, c in mapping_graph.query(query):
        blanknode = rdflib.BNode()
        mapping_graph.add((tm, rdflib.term.URIRef(R2RML_PREDICATE_OBJECT_MAP), blanknode))
        mapping_graph.add((blanknode, rdflib.term.URIRef(R2RML_OBJECT_CONSTANT_SHORTCUT), c))
        mapping_graph.add((blanknode, rdflib.term.URIRef(R2RML_PREDICATE_CONSTANT_SHORTCUT), rdflib.RDF.type))

    mapping_graph.remove((None, rdflib.term.URIRef(R2RML_CLASS), None))

    return mapping_graph


def _subject_graph_maps_to_pom(mapping_graph):
    """
    Move graph maps in subject maps to the predicate object maps of subject maps.
    """

    # add the graph maps in the subject maps to every predicate object map of the subject maps
    query = 'SELECT ?sm ?gm ?pom WHERE { ' \
            '?tm <' + R2RML_SUBJECT_MAP + '> ?sm . ' \
            '?sm <' + R2RML_GRAPH_MAP + '> ?gm . ' \
            '?tm <' + R2RML_PREDICATE_OBJECT_MAP + '> ?pom . }'
    for sm, gm, pom in mapping_graph.query(query):
        mapping_graph.add((pom, rdflib.term.URIRef(R2RML_GRAPH_MAP), gm))

    # remove the graph maps from the subject maps
    query = 'SELECT ?sm ?gm WHERE { ' \
            '?tm <' + R2RML_SUBJECT_MAP + '> ?sm . ' \
            '?sm <' + R2RML_GRAPH_MAP + '> ?gm . }'
    for sm, gm in mapping_graph.query(query):
        mapping_graph.remove((sm, rdflib.term.URIRef(R2RML_GRAPH_MAP), gm))

    return mapping_graph


def _complete_pom_with_default_graph(mapping_graph):
    """
    Complete predicate object maps without graph maps with rr:defaultGraph.
    """

    query = 'SELECT DISTINCT ?tm ?pom WHERE { ' \
            '?tm <' + R2RML_PREDICATE_OBJECT_MAP + '> ?pom . ' \
            'OPTIONAL { ?pom <' + R2RML_GRAPH_MAP + '> ?gm . } . ' \
            'FILTER ( !bound(?gm) ) }'
    for tm, pom in mapping_graph.query(query):
        blanknode = rdflib.BNode()
        mapping_graph.add((pom, rdflib.term.URIRef(R2RML_GRAPH_MAP), blanknode))
        mapping_graph.add((blanknode, rdflib.term.URIRef(R2RML_CONSTANT), rdflib.term.URIRef(R2RML_DEFAULT_GRAPH)))

    return mapping_graph


def _complete_termtypes(mapping_graph):
    """
    Completes term types of mapping rules that do not have rr:termType property as indicated in R2RML specification
    (https://www.w3.org/2001/sw/rdb2rdf/r2rml/#termtype).
    """

    # add missing blanknode termtypes in the constant-valued object maps
    query = 'SELECT DISTINCT ?term_map ?constant WHERE { ' \
            '?term_map <' + R2RML_CONSTANT + '> ?constant . ' \
            'OPTIONAL { ?term_map <' + R2RML_TERM_TYPE + '> ?termtype . } . ' \
            'FILTER ( !bound(?termtype) && isBlank(?constant) ) }'
    for term_map, _ in mapping_graph.query(query):
        mapping_graph.add((term_map, rdflib.term.URIRef(R2RML_TERM_TYPE), rdflib.term.URIRef(R2RML_BLANK_NODE)))

    # add missing literal termtypes in the constant-valued object maps
    query = 'SELECT DISTINCT ?term_map ?constant WHERE { ' \
            '?term_map <' + R2RML_CONSTANT + '> ?constant . ' \
            'OPTIONAL { ?term_map <' + R2RML_TERM_TYPE + '> ?termtype . } . ' \
            'FILTER ( !bound(?termtype) && isLiteral(?constant) ) }'
    for term_map, _ in mapping_graph.query(query):
        mapping_graph.add((term_map, rdflib.term.URIRef(R2RML_TERM_TYPE), rdflib.term.URIRef(R2RML_LITERAL)))

    # add missing literal termtypes in the object maps
    query = 'SELECT DISTINCT ?om ?pom WHERE { ' \
            '?pom <' + R2RML_OBJECT_MAP + '> ?om . ' \
            'OPTIONAL { ?om <' + R2RML_TERM_TYPE + '> ?termtype . } . ' \
            'OPTIONAL { ?om <' + RML_REFERENCE + '> ?column . } . ' \
            'OPTIONAL { ?om <' + R2RML_LANGUAGE + '> ?language . } . ' \
            'OPTIONAL { ?om <' + R2RML_DATATYPE + '> ?datatype . } . ' \
            'FILTER ( !bound(?termtype) && ( bound(?column) || bound(?language) || bound(?datatype) ) ) }'
    for om, _ in mapping_graph.query(query):
        mapping_graph.add((om, rdflib.term.URIRef(R2RML_TERM_TYPE), rdflib.term.URIRef(R2RML_LITERAL)))

    # now all missing termtypes are IRIs
    for term_map_property in [R2RML_SUBJECT_MAP, R2RML_PREDICATE_MAP, R2RML_OBJECT_MAP, R2RML_GRAPH_MAP]:
        query = 'SELECT DISTINCT ?term_map ?x WHERE { ' \
                '?x <' + term_map_property + '> ?term_map . ' \
                'OPTIONAL { ?term_map <' + R2RML_TERM_TYPE + '> ?termtype . } . ' \
                'FILTER ( !bound(?termtype) ) }'
        for term_map, _ in mapping_graph.query(query):
            mapping_graph.add((term_map, rdflib.term.URIRef(R2RML_TERM_TYPE), rdflib.term.URIRef(R2RML_IRI)))

    return mapping_graph


def _remove_string_datatypes(mapping_graph):
    """
    Removes xsd:string data types. xsd:string is equivalent to not specifying any data type.
    """

    mapping_graph.remove((None, rdflib.term.URIRef(R2RML_DATATYPE), rdflib.term.URIRef(XSD_STRING)))

    return mapping_graph


def _complete_rml_classes(mapping_graph):
    """
    TODO
    """

    return mapping_graph


def _get_join_object_maps_join_conditions(join_query_results):
    """
    Creates a dictionary with the results of the JOIN_CONDITION_PARSING_QUERY. The keys are the identifiers of the
    child triples maps of the join condition. The values of the dictionary are in turn other dictionaries with two
    items, child_value and parent_value, representing a join condition.
    """

    join_conditions_dict = {}

    for join_condition in join_query_results:
        # add the child triples map identifier if it is not in the dictionary
        if join_condition.object_map not in join_conditions_dict:
            join_conditions_dict[join_condition.object_map] = {}

        # add the new join condition (note that several join conditions can apply in a join)
        join_conditions_dict[join_condition.object_map][str(join_condition.join_condition)] = \
            {'child_value': str(join_condition.child_value), 'parent_value': str(join_condition.parent_value)}

    return join_conditions_dict


def _transform_mappings_into_dataframe(mapping_query_results, join_query_results, section_name):
    """
    Builds a Pandas DataFrame from the results obtained from MAPPING_PARSING_QUERY and
    JOIN_CONDITION_PARSING_QUERY for one source.
    """

    # mapping rules in graph to DataFrame
    source_mappings_df = pd.DataFrame(mapping_query_results.bindings)
    source_mappings_df.columns = source_mappings_df.columns.map(str)

    # process mapping rules with joins
    # create a dict with child triples maps in the keys and its join conditions in the values
    join_conditions_dict = _get_join_object_maps_join_conditions(join_query_results)
    # map the dict with the join conditions to the mapping rules in the DataFrame
    source_mappings_df['join_conditions'] = source_mappings_df['object_map'].map(join_conditions_dict)
    # needed for later hashing the dataframe
    source_mappings_df['join_conditions'] = source_mappings_df['join_conditions'].where(
        pd.notna(source_mappings_df['join_conditions']), '')
    # convert the join condition dicts to string (can later be converted back to dict)
    source_mappings_df['join_conditions'] = source_mappings_df['join_conditions'].astype(str)
    # object_map column no longer needed, remove it
    source_mappings_df = source_mappings_df.drop('object_map', axis=1)

    # link the mapping rules to their data source name
    source_mappings_df['source_name'] = section_name

    return source_mappings_df


def _is_delimited_identifier(identifier):
    """
    Checks if an identifier is delimited or not.
    """

    if len(identifier) > 2:
        if identifier[0] == '"' and identifier[len(identifier) - 1] == '"':
            return True
    return False


def _get_undelimited_identifier(identifier):
    """
    Removes delimiters from the identifier if it is delimited.
    """

    if pd.notna(identifier):
        identifier = str(identifier)
        if _is_delimited_identifier(identifier):
            return identifier[1:-1]
    return identifier


def _get_valid_template_identifiers(template):
    """
    Removes delimiters from delimited identifiers in a template.
    """

    if pd.notna(template):
        return template.replace('{"', '{').replace('"}', '}')
    return template


class MappingParser:

    def __init__(self, config):
        self.mappings_df = pd.DataFrame(columns=MAPPINGS_DATAFRAME_COLUMNS)
        self.config = config

    def __str__(self):
        return str(self.mappings_df)

    def __repr__(self):
        return repr(self.mappings_df)

    def __len__(self):
        return len(self.mappings_df)

    def parse_mappings(self):
        self._get_from_r2_rml()
        self._preprocess_mappings()
        self._infer_datatypes()

        self.validate_mappings()

        logging.info(str(len(self.mappings_df)) + ' mapping rules retrieved.')

        # generate mapping partitions
        mapping_partitioner = MappingPartitioner(self.mappings_df, self.config)
        self.mappings_df = mapping_partitioner.partition_mappings()

        # replace empty strings with NaN
        self.mappings_df = self.mappings_df.replace(r'^\s*$', np.nan, regex=True)

        return self.mappings_df

    def _get_from_r2_rml(self):
        """
        Parses the mapping files of all data sources in the config file and adds the parsed mappings rules to a
        common DataFrame for all data sources. If parallelization is enabled and multiple data sources are provided,
        each mapping file is parsed in parallel.
        """

        if self.config.is_multiprocessing_enabled() and self.config.has_multiple_data_sources():
            pool = mp.Pool(self.config.get_number_of_processes())
            mappings_dfs = pool.map(self._parse_data_source_mapping_files, self.config.get_data_sources_sections())
            self.mappings_df = pd.concat([self.mappings_df, pd.concat(mappings_dfs)])
        else:
            for section_name in self.config.get_data_sources_sections():
                data_source_mappings_df = self._parse_data_source_mapping_files(section_name)
                self.mappings_df = pd.concat([self.mappings_df, data_source_mappings_df])

        self.mappings_df = self.mappings_df.reset_index(drop=True)

    def _parse_data_source_mapping_files(self, section_name):
        """
        Creates a Pandas DataFrame with the mapping rules of a data source. It loads the mapping files in an rdflib
        graph and recognizes the mapping language used. Mappings are translated to RML.
        It performs queries MAPPING_PARSING_QUERY and JOIN_CONDITION_PARSING_QUERY and process the results to build a
        DataFrame with the mapping rules. Also verifies that there are not repeated triples maps in the mappings.
        """

        # create an empty graph
        mapping_graph = rdflib.Graph()

        mapping_file_paths = self.config.get_mappings_files(section_name)
        try:
            # load mapping rules to the graph
            [mapping_graph.parse(f, format=rdflib.util.guess_format(f)) for f in mapping_file_paths]
        except Exception as n3_mapping_parse_exception:
            raise Exception(n3_mapping_parse_exception)

        # convert R2RML rules to RML, so that we can assume RML for parsing
        mapping_graph = _mapping_to_rml(mapping_graph)
        # convert rr:class to new POMs
        mapping_graph = _rdf_class_to_pom(mapping_graph)
        # expand constant shortcut properties rr:subject, rr:predicate, rr:object and rr:graph
        mapping_graph = _expand_constant_shortcut_properties(mapping_graph)
        # move graph maps in subject maps to the predicate object maps of subject maps
        mapping_graph = _subject_graph_maps_to_pom(mapping_graph)
        # complete predicate object maps without graph maps with rr:defaultGraph
        mapping_graph = _complete_pom_with_default_graph(mapping_graph)
        # if a term as no associated rr:termType, complete it according to R2RML specification
        mapping_graph = _complete_termtypes(mapping_graph)
        # remove xsd:string data types as it is equivalent to not specifying any data type
        mapping_graph = _remove_string_datatypes(mapping_graph)
        # add rdf:type RML classes
        mapping_graph = _complete_rml_classes(mapping_graph)

        # parse the mappings with the parsing queries
        mapping_query_results = mapping_graph.query(MAPPING_PARSING_QUERY)
        join_query_results = mapping_graph.query(JOIN_CONDITION_PARSING_QUERY)

        # convert the SPARQL result set with the parsed mappings to DataFrame
        return _transform_mappings_into_dataframe(mapping_query_results, join_query_results, section_name)

    def _preprocess_mappings(self):
        # start by removing duplicated triples
        self.mappings_df = self.mappings_df.drop_duplicates()

        # complete rml:source with file_paths specified in the config file
        self._complete_rml_source_with_config_file_paths()

        # complete source type with config parameters and data file extensions
        self._complete_source_types()

        # ignore the delimited identifiers (this is not conformant with R2MRL specification)
        self._remove_delimiters_from_mappings()

        # create a unique id for each mapping rule
        self.mappings_df.insert(0, 'id', self.mappings_df.reset_index(drop=True).index)

        self._remove_self_joins_from_mappings()

        # TODO: this is not correct as this rules could be referenced by a join with join conditions
        # remove mapping rules with no predicate or object (subject map is conserved because rdf class was added as POM)
        self.mappings_df = self.mappings_df.dropna(subset=['predicate_constant', 'predicate_template',
                                                           'predicate_reference', 'object_constant', 'object_template',
                                                           'object_reference'], how='all')

    def _complete_source_types(self):
        """
        Adds a column with the source type. The source type is inferred for RDB through the parameter db_url provided
        in the mapping file. For data files the source type is inferred from the file extension.
        """

        for i, mapping_rule in self.mappings_df.iterrows():
            if self.config.has_database_url(mapping_rule['source_name']):
                self.mappings_df.at[i, 'source_type'] = RDB
            elif pd.notna(mapping_rule['data_source']):
                file_extension = os.path.splitext(str(mapping_rule['data_source']))[1][1:].strip()
                self.mappings_df.at[i, 'source_type'] = file_extension.upper()
            else:
                raise Exception('No source type could be retrieved for mapping rule some mapping rules.')

    def _complete_rml_source_with_config_file_paths(self):
        """
        Overrides rml:source in the mappings with the file_path parameter in the config file for each data source
        section if provided.
        """

        for section_name in self.config.get_data_sources_sections():
            if self.config.has_file_path(section_name):
                self.mappings_df.loc[self.mappings_df['source_name'] == section_name, 'data_source'] = \
                    self.config.get_file_path(section_name)

    def _remove_delimiters_from_mappings(self):
        """
        Removes delimiters from all identifiers in the mapping rules in the input DataFrame.
        """

        for i, mapping_rule in self.mappings_df.iterrows():
            self.mappings_df.at[i, 'tablename'] = _get_undelimited_identifier(mapping_rule['tablename'])
            self.mappings_df.at[i, 'subject_template'] = _get_valid_template_identifiers(
                mapping_rule['subject_template'])
            self.mappings_df.at[i, 'subject_reference'] = _get_undelimited_identifier(
                mapping_rule['subject_reference'])
            self.mappings_df.at[i, 'graph_reference'] = _get_undelimited_identifier(
                mapping_rule['graph_reference'])
            self.mappings_df.at[i, 'graph_template'] = _get_valid_template_identifiers(
                mapping_rule['graph_template'])
            self.mappings_df.at[i, 'predicate_template'] = _get_valid_template_identifiers(
                mapping_rule['predicate_template'])
            self.mappings_df.at[i, 'predicate_reference'] = _get_undelimited_identifier(
                mapping_rule['predicate_reference'])
            self.mappings_df.at[i, 'object_template'] = _get_valid_template_identifiers(
                mapping_rule['object_template'])
            self.mappings_df.at[i, 'object_reference'] = _get_undelimited_identifier(
                mapping_rule['object_reference'])

            # if join_condition is not null and it is not empty
            if pd.notna(mapping_rule['join_conditions']) and mapping_rule['join_conditions']:
                join_conditions = eval(mapping_rule['join_conditions'])
                for key, value in join_conditions.items():
                    join_conditions[key]['child_value'] = _get_undelimited_identifier(
                        join_conditions[key]['child_value'])
                    join_conditions[key]['parent_value'] = _get_undelimited_identifier(
                        join_conditions[key]['parent_value'])
                    self.mappings_df.at[i, 'join_conditions'] = str(join_conditions)

    def _remove_self_joins_from_mappings(self):

        # because RDB has no rml:source, use source_name for rml:source to facilitate self-join elimination
        self.mappings_df.loc[self.mappings_df['source_type'] == RDB, 'data_source'] = self.mappings_df['source_name']

        for i, mapping_rule in self.mappings_df.iterrows():
            if pd.notna(mapping_rule['object_parent_triples_map']):
                parent_triples_map_rule = get_mapping_rule_from_triples_map_id(self.mappings_df, mapping_rule[
                    'object_parent_triples_map'])

                # str() is to be able to compare np.nan
                if str(mapping_rule['data_source']) == str(parent_triples_map_rule['data_source']) and \
                        str(mapping_rule['tablename']) == str(parent_triples_map_rule['tablename']) and \
                        str(mapping_rule['iterator']) == str(parent_triples_map_rule['iterator']) and \
                        str(mapping_rule['query']) == str(parent_triples_map_rule['query']):

                    remove_join = True
                    # check that all conditions in the join condition have the same references
                    try:
                        join_conditions = eval(mapping_rule['join_conditions'])
                        for key, join_condition in join_conditions.items():
                            if join_condition['child_value'] != join_condition['parent_value']:
                                remove_join = False
                    except:
                        # eval() has failed because the are no join conditions, the join can be removed
                        remove_join = True

                    if remove_join:
                        self.mappings_df.at[i, 'object_parent_triples_map'] = ''
                        self.mappings_df.at[i, 'join_conditions'] = ''
                        self.mappings_df.at[i, 'object_constant'] = parent_triples_map_rule.at['subject_constant']
                        self.mappings_df.at[i, 'object_template'] = parent_triples_map_rule.at['subject_template']
                        self.mappings_df.at[i, 'object_reference'] = parent_triples_map_rule.at['subject_reference']
                        self.mappings_df.at[i, 'object_termtype'] = parent_triples_map_rule.at['subject_termtype']

                        logging.debug('Removed self-join from mapping rule `' + str(mapping_rule['id']) + '`.')

    def _infer_datatypes(self):
        """
        Get RDF datatypes for rules corresponding to relational data sources if they are not overridden in the mapping
        rules. The inferring of RDF datatypes is defined in R2RML specification
        (https://www.w3.org/2001/sw/rdb2rdf/r2rml/#natural-mapping).
        """

        # return if datatype inferring is disabled in the config
        if not self.config.infer_sql_datatypes():
            return

        for i, mapping_rule in self.mappings_df.iterrows():
            # datatype inference only applies to relational data sources
            if (mapping_rule['source_type'] == RDB) and (
                    # datatype inference only applies to literals
                    str(mapping_rule['object_termtype']) == R2RML_LITERAL) and (
                    # if the literal has a language tag or an overridden datatype, datatype inference does not apply
                    pd.isna(mapping_rule['object_datatype']) and pd.isna(mapping_rule['object_language'])):

                if pd.notna(mapping_rule['object_reference']):
                    inferred_data_type = get_rdb_reference_datatype(self.config, mapping_rule,
                                                                    mapping_rule['object_reference'])

                    if not inferred_data_type:
                        # no data type was inferred
                        continue

                    self.mappings_df.at[i, 'object_datatype'] = inferred_data_type
                    if pd.notna(mapping_rule['tablename']):
                        logging.debug("`" + inferred_data_type + "` datatype inferred for column `" +
                                      mapping_rule['object_reference'] + "` of table `" +
                                      mapping_rule['tablename'] + "` in data source `" +
                                      mapping_rule['source_name'] + "`.")
                    elif pd.notna(mapping_rule['query']):
                        logging.debug("`" + inferred_data_type + "` datatype inferred for reference `" +
                                      mapping_rule['object_reference'] + "` in query [" +
                                      mapping_rule['query'] + "] in data source `" +
                                      mapping_rule['source_name'] + "`.")

    def validate_mappings(self):
        """
        Checks that the mapping rules in the input DataFrame are valid. If something is wrong in the mappings the
        execution is stopped. Specifically it is checked that termtypes are correct, and that language tags and
        datatypes are used properly. Also checks that different data sources do not have triples map with the same id.
        """

        # check termtypes are correct (i.e. that they are rr:IRI, rr:BlankNode or rr:Literal and that subject map is
        # not a rr:literal). Use subset operation
        subject_termtypes = set([str(termtype) for termtype in set(self.mappings_df['subject_termtype'])])
        if not (subject_termtypes <= {R2RML_IRI, R2RML_BLANK_NODE}):
            raise ValueError('Found an invalid subject termtype. Found values ' + str(subject_termtypes) +
                             '. Subject maps must be ' + R2RML_IRI + ' or ' + R2RML_BLANK_NODE + '.')

        object_termtypes = set([str(termtype) for termtype in set(self.mappings_df['object_termtype'])])
        if not (object_termtypes <= {R2RML_IRI, R2RML_BLANK_NODE, R2RML_LITERAL}):
            raise ValueError('Found an invalid object termtype. Found values ' + str(object_termtypes) +
                             '. Object maps must be ' + R2RML_IRI + ', ' + R2RML_BLANK_NODE + ' or ' +
                             R2RML_LITERAL + '.')

        # if there is a datatype or language tag then the object map termtype must be a rr:Literal
        if len(self.mappings_df.loc[(self.mappings_df['object_termtype'] != R2RML_LITERAL) &
                                    pd.notna(self.mappings_df['object_datatype']) &
                                    pd.notna(self.mappings_df['object_language'])]) > 0:
            raise Exception('Found object maps with a language tag or a datatype, '
                            'but that do not have termtype rr:Literal.')

        # language tags and datatypes cannot be used simultaneously, language tags are used if both are given
        if len(self.mappings_df.loc[pd.notna(self.mappings_df['object_language']) &
                                    pd.notna(self.mappings_df['object_datatype'])]) > 0:
            logging.warning('Found object maps with a language tag and a datatype. Both of them cannot be used '
                            'simultaneously for the same object map, and the language tag has preference.')

        # check that language tags are valid
        language_tags = set(self.mappings_df['object_language'].dropna())
        # the place to look for valid language subtags is the IANA Language Subtag Registry
        # (https://www.iana.org/assignments/language-subtag-registry/language-subtag-registry)
        for language_tag in language_tags:
            # in general, if the language subtag is longer than 3 characters it is not valid
            if len(language_tag.split('-')[0]) > 3:
                raise ValueError('Found invalid language tag `' + language_tag +
                                 '`. Language tags must be in the IANA Language Subtag Registry.')

        # check that a triples map id is not repeated in different data sources
        # Get unique source names and triples map identifiers
        aux_mappings_df = self.mappings_df[['source_name', 'triples_map_id']].drop_duplicates()
        # get repeated triples map identifiers
        repeated_triples_map_ids = get_repeated_elements_in_list(list(aux_mappings_df['triples_map_id'].astype(str)))
        # of those repeated identifiers
        repeated_triples_map_ids = [tm_id for tm_id in repeated_triples_map_ids]
        if len(repeated_triples_map_ids) > 0:
            raise Exception('The following triples maps appear in more than one data source: ' +
                            str(repeated_triples_map_ids) +
                            '. Check the mapping files, one triple map cannot be repeated in different data sources.')



import os
import re
import sys

from setuptools import setup, find_packages


CURRENT_PYTHON = sys.version_info[:2]
REQUIRED_PYTHON = (3, 7)

# borrowed from urllib3
if CURRENT_PYTHON < REQUIRED_PYTHON:
    sys.stderr.write("""
==========================
Unsupported Python version
==========================
This version of Morph-KGC requires Python {}.{}, but you're trying to install it on Python {}.{}.
""".format(*(REQUIRED_PYTHON + CURRENT_PYTHON)))
    sys.exit(1)


# borrowed from SQLAlchemy
with open(os.path.join(os.path.dirname(__file__), 'src', 'morph_kgc', '_version.py')) as file:
    version = (re.compile(r""".*__version__ = ["'](.*?)['"]""", re.S).match(file.read()).group(1))


with open(os.path.join(os.path.dirname(__file__), 'README.md'), 'r', encoding='utf-8') as file:
    readme = file.read()
    # remove morph logo
    readme = re.sub("<p[^>]*>", "", readme)
    readme = re.sub("<img[^>]*>", "", readme)
    readme = re.sub("</?p[^>]*>", "", readme)


with open(os.path.join(os.path.dirname(__file__), 'requirements.txt')) as file:
    requirements = [line.strip().replace(' ', '') for line in file.readlines()]


setup(
    name='morph_kgc',
    version=version,
    author='Julián Arenas-Guerrero',
    author_email='arenas.guerrero.julian@outlook.com',
    license='Apache 2.0',
    description='Scalable [R2]RML engine to create RDF knowledge graphs from heterogeneous data sources.',
    keywords='morph-kgc, rdf, r2rml, rml, knowledge graphs, data integration'
    ,
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/oeg-upm/Morph-KGC',
    project_urls={
        'Documentation': 'https://github.com/oeg-upm/Morph-KGC/wiki',
        'Source code': 'https://github.com/oeg-upm/Morph-KGC',
        'Issue tracker': 'https://github.com/oeg-upm/Morph-KGC/issues',
    },
    include_package_data=True,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Pre-processors',
        'Topic :: Database',
        'Topic :: Utilities',
        'Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator'
    ],
    install_requires=requirements,
    python_requires='>=3.7, <4',
)



__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "arenas.guerrero.julian@outlook.com"


import sys
import pandas as pd
import logging
import time

from .utils import get_delta_time
from .mapping.mapping_parser import MappingParser
from .materializer import Materializer


def retrieve_mappings(config):
    if config.is_read_parsed_mappings_file_provided():
        # retrieve parsed mapping from file and finish mapping processing
        mappings = pd.read_csv(config.get_parsed_mappings_read_path(), keep_default_na=False)
        logging.info(str(len(mappings)) + ' mappings rules loaded from file.')
    else:
        mappings_parser = MappingParser(config)

        start_time = time.time()
        mappings = mappings_parser.parse_mappings()
        logging.info('Mappings processed in ' + get_delta_time(start_time) + ' seconds.')

    if config.is_write_parsed_mappings_file_provided():
        mappings.sort_values(by=['id'], axis=0).to_csv(config.get_parsed_mappings_write_path(), index=False)
        logging.info('Parsed mapping rules saved to file.')
        sys.exit()

    return mappings


def process_materialization(mappings, config):
    materializer = Materializer(mappings, config)

    start_time = time.time()
    if config.is_multiprocessing_enabled():
        materializer.materialize_concurrently()
    else:
        materializer.materialize()

    logging.info('Materialization finished in ' + get_delta_time(start_time) + ' seconds.')



__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "arenas.guerrero.julian@outlook.com"


import argparse
import os

from configparser import ExtendedInterpolation

from .utils import configure_logger
from .constants import __copyright__
from .config import Config
from ._version import __version__


def _existing_file_path(file_path):
    """
    Checks whether a file exists.
    """

    file_path = str(file_path).strip()
    if not os.path.isfile(file_path):
        raise argparse.ArgumentTypeError("%r is not a valid file path." % file_path)

    return file_path


def _parse_arguments():
    """
    Parses command line arguments.
    """

    parser = argparse.ArgumentParser(
        description='Generate Knowledge Graphs from Heterogeneous Data Sources.',
        epilog=__copyright__,
        allow_abbrev=False,
        prog='python3 -m morph_kgc',
        argument_default=argparse.SUPPRESS
    )

    parser.add_argument('config', type=_existing_file_path, help='path to the configuration file')
    parser.add_argument('-v', '--version', action='version', version='Morph-KGC ' + __version__ + ' | ' + __copyright__)

    return parser.parse_args()


def _parse_config(config):
    """
    Parses the config file. Logger is configured.
    """

    config.complete_configuration_with_defaults()

    config.validate_configuration_section()

    configure_logger(config.get_logging_level(), config.get_logging_file())
    config.log_config_info()

    return config


def load_config_from_command_line():
    """
    Parses command line arguments.
    """

    args = _parse_arguments()

    config = Config(interpolation=ExtendedInterpolation())
    config.read(args.config)

    config = _parse_config(config)

    return config


def load_config_from_argument(config_entry):
    """
    Parses a config argument. It can be a file path or a string.
    """

    config = Config(interpolation=ExtendedInterpolation())
    if os.path.isfile(config_entry):
        config.read(config_entry)
    else:
        # it is a string
        config.read_string(config_entry)

    config = _parse_config(config)

    return config
